import os

from flask import current_app

__all__ = ['get_app_path', 'get_app_config']


def get_app_path(*path):
    """
    Get the application path.
    app.get_app_path("_log", "sys.log")
    :param path:    -The relative path
    :return:        -The path relative to the application root directory
    """
    return os.path.join(os.getcwd(), *path)


def get_app_config(key, default=None):
    """
    Set the specified config value of the current app
    :param key:
    :param default:
    :return:
    """
    return current_app.config.get(key, default)



from contextlib import contextmanager

from flask import g

from . import DBSession
from ._model import ModelMixin
from ..utils import get_g_cache, set_g_cache

__all__ = ['create_instance', 'create_relationships',
           'query_multiple_model',
           'append_debug_queries', 'get_debug_queries',
           'get_db_session', 'db_session', 'close_db_session',
           'model_to_dict']


def create_instance(model_cls, data, create_relationship=True):
    """
    Create an instance of the specified model class with the data
    :param model_cls: The specified model class.
    :param data: The data used to create instance.
    :param create_relationship: If true, the relationships will be created.
    :return:
    """
    if not isinstance(data, dict):
        return None

    col_value_dict = model_cls.get_columns_json(data)
    instance = model_cls(**col_value_dict)

    if create_relationship is True:
        relationships = create_relationships(model_cls, data)
        for key in relationships:
            setattr(instance, key, relationships[key])
    return instance


def create_relationships(model_cls, data):
    """
    Create the relationship dict of the specified model class with the data
    :param model_cls:
    :param data:
    :return:
    """
    relationships = model_cls.get_relationships()
    relationship_map = {}
    for key in relationships.keys():
        relationship_cls = relationships[key].mapper.class_
        relationship_kwargs = data.get(key)
        if isinstance(relationship_kwargs, list):  # 1:n
            relationship = []
            for item in relationship_kwargs:
                r_ins = create_instance(relationship_cls, item)
                if r_ins is not None:
                    relationship.append(r_ins)
        else:
            relationship = create_instance(relationship_cls, relationship_kwargs)  # 1:1

        if relationship is not None:
            relationship_map[key] = relationship

    return relationship_map


def query_multiple_model(*cls_list):
    """
    Query all the data of the multiple specified model class.
    :param cls_list:
    :return:If failed, returns the failed tuple, otherwise, returns the the data list.
    """
    result = []
    for cls in cls_list:
        r = cls.query_all()

        if type(r) is tuple:  # ModelMixin
            if r[0] is not True:  # error
                return r
            else:
                result.append(r[1])
        else:
            result.append(r)
    return result


def append_debug_queries(query):
    debug_queries = getattr(g, 'z_debug_queries', None)
    if debug_queries is None:
        g.z_debug_queries = debug_queries = []
    debug_queries.append(query)


def get_debug_queries():
    return getattr(g, 'z_debug_queries', [])


def get_db_session():
    """
    Get the db session from g.
    If not exist, create a session and return.
    :return:
    """
    session = get_g_cache('db_session')
    if session is None:
        session = DBSession()
        set_g_cache('db_session', session)

    return session


def close_db_session():
    session = get_g_cache('db_session')
    if session is not None:
        session.close()


@contextmanager
def db_session(do_commit=True):
    """
    Database session context manager.

    instance = create_instance(cls, json_data)
    with db_session() as session:
        session.add(instance)

    :param do_commit: If false, session will not commit,generally used for query operations
    :return:
    """
    session = get_db_session()
    try:
        yield session
        if do_commit is not False:
            session.commit()
    except Exception as e:
        if do_commit is not False:
            session.rollback()
        raise e


def model_to_dict(ins, option=None):
    """
    Convert model data to json dict

    :param ins:
    :param option:
    :return:
    """
    if isinstance(ins, list):
        data_list = []
        for item in ins:
            if isinstance(item, ModelMixin):
                data_list.append(item.to_dict(option))
            else:
                data_list.append(item)
        return data_list
    elif isinstance(ins, ModelMixin):
        return ins.to_dict(option)
    else:
        return ins



from sqlalchemy import text

from .. import res_status_codes
from ..utils import find_list, filter_list, is_str, is_dict, ins_to_dict

__all__ = ['BaseModelMixin']


class BaseModelMixin:
    # -------------------------------------------about model-------------------------------------------
    @classmethod
    def get_class_name(cls):
        """
        Get the name of the model class.
        :return:
        """
        return cls.__name__

    @classmethod
    def get_columns(cls):
        """
        Get all the columns of the model class.
        Menu.get_cls_columns()
        :return:
        """
        return cls.__table__.columns

    @classmethod
    def get_column_field(cls, col):
        """
        Get the field variable name of the column, default return the key of the column
        Used to get json fields to create or update data(get_columns_json), or convert data to json(_to_json)

        If the field name is different from the database column name
        #1 set field in info dict.(recommend)
        system_default = Column('default', Boolean, default=False, info={'field': 'system_default'})

        #2 rewrite to get the field name
        system_default = Column('default', Boolean, default=False)
        @classmethod
        def get_column_field(cls, col):
            key = col.key
            if key == 'default':
                return 'system_default'
            return key

        :param col:
        :return:
        """
        field = col.info.get('field')
        if field:
            return field
        return col.key

    @classmethod
    def get_primary_column(cls):
        """
         Get the primary column of the model class.
        :return:
        """
        return find_list(cls.get_columns(), lambda c: c.primary_key is True)

    @classmethod
    def get_primary_key(cls):
        """
        Get the primary_key of the model class.
        :return:
        """
        col = cls.get_primary_column()
        if col is not None:  # must
            return col.key
        return None

    @classmethod
    def get_primary_field(cls):
        """
        Get the primary_key field of the model class.
        :return:
        """
        col = cls.get_primary_column()
        if col is not None:  # must
            return cls.get_column_field(col)  # col.key
        return None

    @classmethod
    def get_unique_columns(cls):
        """
        Get the unique column list of the model class
         Menu.get_cls_unique_columns()
        :return:
        """
        return filter_list(cls.get_columns(), lambda col: col.unique is True)

    @classmethod
    def get_relationships(cls):
        """
        Get all the relationship list of the model class.
         Menu.get_cls_relationships()
        :return:
        """
        return cls.__mapper__.relationships

    # -------------------------------------------About data-------------------------------------------
    def to_dict(self, option=None):
        """
        # 过滤
            -不同的class会有不同的条件过滤-->过滤掉不需要的属性(create_user/password)
            -过滤条件可以是class上的属性，也可以是传递到方法中的exclude
                -如果有include--->只返回include中的属性
                -如果没有include而有exclude--->将exclude中的属性过滤掉
        *# 如果属性是关系对象/列表，则
        *#
        *#
        *#
        # 单个对象很容易处理，但是 relationships不好处理
        # 不同的relationship，cascade不同
        # relation中可能有嵌套
        # 有可能两个对象里都有同一个对象
        Convert model data to dict

        ins_to_dict(A, {
            'cascade': 3,  # 如果子项没有cascade，则使用父项-1，如果cascade不大于0，则不对象属性
            'recursion_value': '{...}'
            # 'include': ['a1'],  # 只有include中的字段才会返回，不区分值是否是对象，include的优先级>exclude
            # 'exclude': ['a2'],  # 只有exclude外的字段才会返回，不区分值是否是对象
            # 'getattr_items': lambda ins, item_cascade, option, path_keys: ins.__dict__.items(),
            #
            # 'bb': {  # 某个子项的设置
            #     # 'cascade': 1,
            #     # 'include': ['b1'],
            #     'exclude': ['b2'],
            # },
            # 'bb.xx': {  # 子项的子项的设置
            #     'exclude': ['x1']
            # },
            # 'bb.yy': {
            #     'exclude': ['y2']
            # },
            # 'cc': {  # cc列表中的每个数据的设置
            #     'exclude': ['n'],
            #     'cascade': 1
            # },
            # "cc.xx": {
            #     'exclude': ['x2']
            # }
        })
        """
        cls = self.__class__
        opt = {
            'getattrs': lambda ins, *args, **kwargs: ins.__class__.get_to_dict_attrs(ins, *args, **kwargs),
            'include': lambda ins, key: ins.__class__.to_dict_field_filter(key),
        }
        if option:
            opt.update(option)
        return ins_to_dict(self, opt)

    @classmethod
    def get_to_dict_attrs(cls, ins, cascade, *args):
        fields = [cls.get_column_field(col) for col in cls.get_columns()]
        if cascade > 0:
            for relationship in cls.get_relationships():
                lazy = relationship.lazy
                if lazy != 'dynamic' and lazy != 'noload':
                    fields.append(relationship.key)
        items = {}
        for field in fields:
            items[field] = getattr(ins, field)
        return items

    @classmethod
    def to_dict_field_filter(cls, field):
        """
        to_dict callback.
        If return false, the field will not be returned.
        :param field:
        :return:
        """
        return True

    @classmethod
    def get_columns_json(cls, data):
        """
        Get the data corresponding to the column.
        :param data::
        :return:
        """
        json = {}
        auto_columns = getattr(cls, 'auto_columns', [])
        auto_fields = []
        for col in auto_columns:
            if is_str(col):
                auto_fields.append(col)
            else:
                auto_fields.append(cls.get_column_field(col))

        for col in cls.get_columns():
            field = cls.get_column_field(col)  # col.key
            if (field in data) and (field not in auto_fields) and (not col.info.get('auto', False)):
                value = data.get(field)
                if col.nullable is False:
                    if not (value is None or (is_str(value) and value.strip() == "")):
                        json[field] = value
                else:
                    json[field] = value

        return json

    # -------------------------------------------add-------------------------------------------
    @classmethod
    def check_add_data(cls, data):
        """
        Check the the added json data.
        --validate the json data
        --check unique
        If the check result is not True, the adding process will be terminated and the check result will be returned to the client.

        :param data:  The data to be added
        :return: True|Error Message
        """
        if not is_dict(data):
            if data:
                return data
            return res_status_codes.bad_request
        return cls._check_unique(data)

    @classmethod
    def add_db(cls, data):
        """
        Add the data to the db.

        :param data:
        :return:
        """
        instance = create_instance(cls, data)
        with db_session() as session:
            session.add(instance)
        return instance

    @classmethod
    def bulk_add(cls, items, with_relationship=False):
        """
        Perform a bulk add of the given list of mapping dictionaries

        :param items:
        :param with_relationship:
        :return:
        """
        if len(items) == 0:
            return
        if with_relationship is True:
            with db_session() as session:
                ins_list = []
                for data_item in items:
                    ins_list.append(create_instance(cls, data_item))
                    # instance = create_instance(cls, data_item)
                    # session.add(instance)
                session.add_all(ins_list)
        else:
            with db_session() as session:
                session.bulk_insert_mappings(cls, items)

    # -------------------------------------------update-------------------------------------------
    @classmethod
    def check_update_data(cls, data):
        """
        Check the the updated json data.
        --validate the json data
        --check exist
        --check unique
        If the check result is not True, the update process will be terminated and the check result will be returned to the client.

        :param data: The updated json data
        :return:
        """
        if not is_dict(data):
            if data:
                return data
            return res_status_codes.bad_request

        exist = cls._check_exist(data)
        if exist is not True:
            return exist
        return cls._check_unique(data)

    @classmethod
    def update_db(cls, data):
        """
        Update the data to the db.
        Only the fields in json will be updated

        :param data:
        :return:
        """
        pk_value = cls._get_pk_value(data)
        instance = cls.query_by_pk(pk_value)
        if instance:
            with db_session() as session:
                # Only the fields in json will be updated
                for field in cls.get_columns_json(data):
                    setattr(instance, field, data.get(field))
                relationships = create_relationships(cls, data)
                for field in relationships:
                    setattr(instance, field, relationships[field])
        return instance

    @classmethod
    def bulk_update(cls, items, with_relationship=False):
        """
        Perform a bulk update of the given list of mapping dictionaries
        :param items:
        :param with_relationship:
        :return:
        """
        if len(items) == 0:
            return

        if with_relationship is True:
            with db_session() as session:
                for item in items:
                    pk_value = cls._get_pk_value(item)
                    instance = cls.query_by_pk(pk_value)
                    for field in cls.get_columns_json(item):
                        setattr(instance, field, item.get(field))
                    relationships = create_relationships(cls, item)
                    for field in relationships:
                        setattr(instance, field, relationships[field])
        else:
            with db_session() as session:
                session.bulk_update_mappings(cls, items)

    # def del_attrs(self, with_relationship=True):
    #     """
    #     Delete all attributes of the instance
    #     todo not null auto
    #     :param ins:
    #     :param with_relationship:
    #     :return:
    #     """
    #     cls = self.__class__
    #     pk_col = cls.get_primary_column()
    #     for col in cls.get_columns():
    #         if col is not pk_col:
    #             field = cls.get_column_field(col)  # col.key
    #             delattr(self, field)
    #     if with_relationship is not True:
    #         relationships = cls.get_relationships()
    #         for key in relationships.keys():
    #             delattr(self, key)

    # -------------------------------------------delete-------------------------------------------
    @classmethod
    def check_delete_data(cls, pk_value):
        """
        Check the the deleted data.
        --validate the deleted data
        --check exist
        If the check result is not True, the delete process will be terminated and the check result will be returned to the client.

        :param pk_value:
        :return:
        """
        if pk_value is None:
            return res_status_codes.db_data_not_found

        primary_field = cls.get_primary_field()
        data = {primary_field: pk_value}

        return cls._check_exist(data)

    @classmethod
    def delete_db(cls, pk_value):
        """
        Delete data from the db.

        :param pk_value:
        :return:
        """
        instance = cls.query_by_pk(pk_value)
        if instance is None:  # Object does not exist
            return res_status_codes.db_data_not_found

        with db_session() as session:
            session.delete(instance)
        return instance

    @classmethod
    def bulk_delete(cls, items):
        """
        Perform a bulk delete of the given list of mapping dictionaries

        :param items:
        :return:
        """
        if len(items) == 0:
            return
        with db_session() as session:
            for item in items:
                if is_dict(item):
                    instance = cls.query_by(item)
                else:
                    instance = cls.query_by_pk(item)
                if instance:
                    session.delete(instance)

    # -------------------------------------------query-------------------------------------------

    @classmethod
    def query(cls):
        """
        Return a 'Query' object corresponding to this class

        query = TemplateModel.query()
        print(query.all())

        :return:
        """
        with db_session(do_commit=False) as session:
            return session.query(cls)

    @classmethod
    def get_query_default_order(cls):
        """
        Get the default order of the query.
        column key, not field

        get_primary_key-->get_primary_column to fix
        sqlalchemy.exc.CompileError: Can't resolve label reference for ORDER BY / GROUP BY / DISTINCT etc.
            Textual SQL expression 'id' should be explicitly declared as text('id')
        :return:
        """
        return cls.get_primary_column()

    @classmethod
    def query_by_unique_key(cls, data):
        """
        Query data by the unique values.
        --If exist,returns the instance.
        --Otherwise,returns None

        :param data:
        :return:
        """
        cols = cls.get_unique_columns()
        if len(cols) == 0:
            return None

        ors = []
        for col in cols:
            field = cls.get_column_field(col)  # col.key
            value = data.get(field)
            if value is not None:  # maybe 0
                ors.append(field + "='" + str(value).strip() + "'")

        if len(ors) == 0:
            return None

        with db_session(do_commit=False) as session:
            query = session.query(cls).filter(text(' OR '.join(ors)))
            instance = query.first()
        return instance

    @classmethod
    def query_by_pk(cls, pk_value):
        """
        Query by pk value.
        :param pk_value:
        :return:
        """
        with db_session(do_commit=False) as session:
            instance = session.query(cls).get(pk_value)
        return instance

    @classmethod
    def query_by(cls, by_dict, return_first=False):
        """
        Query by dict object.
        -If first is True, return the first row object or None.
        -If first is not True, return the list result of the query.
        """
        with db_session(do_commit=False) as session:
            if return_first is True:
                result = session.query(cls).filter_by(**by_dict).limit(1).first()
            else:
                result = session.query(cls).filter_by(**by_dict).all()
        return result

    @classmethod
    def query_all(cls):
        """
        Query all the data of the model class.
        :return:
        """
        query_order = cls.get_query_default_order()
        with db_session(do_commit=False) as session:
            query = session.query(cls)
            if query_order is not None:
                query = query.order_by(query_order)
            result = query.all()
        return result

    @classmethod
    def query_pss(cls, pss_option):
        """
        Query data by search, pagination and sort condition.

        SELECT templates.id AS templates_id, templates.name AS templates_name, templates.age AS templates_age, templates.email AS templates_email,
                templates.description AS templates_description, templates.created_at AS templates_created_at, templates.updated_at AS templates_updated_at
        FROM templates
        WHERE (name like '%t%' OR description like '%t%') AND (age>1 AND age<20) AND (email='t1@focus-ui.com' OR email='t2@focus-ui.com') ORDER BY templates.name ASC
         LIMIT ? OFFSET ?
        (20, 0)

        :param pss_option:
        :return:
        """
        filter_likes = pss_option.get('filter_likes', [])
        filter_ands = pss_option.get('filter_ands', [])
        filter_ors = pss_option.get('filter_ors', [])
        offset = pss_option.get('offset', 0)
        limit = pss_option.get('limit', 0)
        # pss_json.get('order') or cls.get_query_default_order()
        # bool(pss_json.get('order')) --> Boolean value of this clause is not defined
        order = pss_option.get('order')
        if order is None:
            order = cls.get_query_default_order()

        with db_session(do_commit=False) as session:
            query = session.query(cls)
            if len(filter_likes) > 0:
                query = query.filter(text('(' + (' OR '.join(filter_likes)) + ')'))
            if len(filter_ands) > 0:
                query = query.filter(text('(' + (' AND '.join(filter_ands)) + ')'))
            if len(filter_ors) > 0:
                query = query.filter(text('(' + (' OR '.join(filter_ors)) + ')'))
            count = query.count()
            if offset < count:
                query = query.order_by(order).offset(offset)
                if limit > 0:
                    query = query.limit(limit)
                items = query.all()
            else:
                items = []
        return {
            'count': count,
            'data': items
        }

    # -------------------------------------------check-------------------------------------------
    @classmethod
    def _get_pk_value(cls, data):
        """
        Get the primary key value from the json data
        :param data:
        :return:
        """
        primary_field = cls.get_primary_field()
        if primary_field:
            return data.get(primary_field)

    @classmethod
    def _check_exist(cls, data):
        """
        Check whether the data exists
        Return True if exists, else return not found code
        :param data:
        :return:
        """
        pk_value = cls._get_pk_value(data)
        if pk_value is not None and cls.query_by_pk(pk_value):
            return True
        return res_status_codes.db_data_not_found  # used to return to the client

    @classmethod
    def _check_unique(cls, data):
        """
        Check whether the data meets uniqueness constraints.
        If meets, returns True, otherwise returns exist code.
        """
        result = cls.query_by_unique_key(data)

        if result is None:
            return True

        primary_field = cls.get_primary_field()
        pk_value = cls._get_pk_value(data)
        if pk_value is not None and pk_value == getattr(result, primary_field):
            return True
        return res_status_codes.db_data_already_exist  # used to return to the client

    def __repr__(self):
        cls = self.__class__
        attrs = []
        for col in cls.get_columns():
            field = cls.get_column_field(col)
            attrs.append(field + '=' + str(getattr(self, field)))
        return cls.get_class_name() + '(' + (', '.join(attrs)) + ')'

    # -------------------------------------------new-------------------------------------------


# must
from ._util import create_instance, create_relationships, db_session



from inspect import isfunction

__all__ = ['Attribute', 'ins_to_dict']


class Attribute:
    """
    Used to create attribute object
    obj = AttrCls(**{'a':1,'b':2})
    obj.c = 3
    obj.a = 10

    print(obj.x)
    print(obj)
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, attr):
        return None

    def __repr__(self):
        t = []
        for key, value in self.__dict__.items():
            t.append(key + '=' + str(value))
        return 'Attribute(' + ", ".join(t) + ')'


def ins_to_dict(ins, option=None):
    """
    Convert instance object to dictionary
    ex)
    ins_to_dict(A, {
        'cascade': 3,  # 如果子项没有cascade，则使用父项-1，如果cascade不大于0，则不对象属性
        'recursion_value': '{...}'
        # 'include': ['a1'],  # 只有include中的字段才会返回，不区分值是否是对象，include的优先级>exclude
        # 'exclude': ['a2'],  # 只有exclude外的字段才会返回，不区分值是否是对象
        # 'getattrs': lambda ins, item_cascade, option, path_keys: ins.__dict__.items(),
        #
        # 'bb': {  # 某个子项的设置
        #     # 'cascade': 1,
        #     # 'include': ['b1'],
        #     'exclude': ['b2'],
        # },
        # 'bb.xx': {  # 子项的子项的设置
        #     'exclude': ['x1']
        # },
        # 'bb.yy': {
        #     'exclude': ['y2']
        # },
        # 'cc': {  # cc列表中的每个数据的设置
        #     'exclude': ['n'],
        #     'cascade': 1
        # },
        # "cc.xx": {
        #     'exclude': ['x2']
        # }
    })
    :param ins:
    :param option:
    :return:
    """
    if isinstance(ins, list):
        result = []
        for item in ins:
            result.append(_ins_to_dict(item, option))
    else:
        result = _ins_to_dict(ins, option)
    return result


def _ins_to_dict(ins, option=None, path_items=None, path_keys=None):
    if path_items is None:
        path_items = []
    else:
        path_items = path_items.copy()
    path_items.append(ins)

    if path_keys is None:
        path_keys = []

    if option is None:
        option = {}

    if len(path_keys) > 0:
        prefix = '.'.join(path_keys)
        item_option = option.get(prefix, {})
    else:
        item_option = option

    if 'cascade' in item_option:
        item_cascade = item_option.get('cascade', 0)
    else:
        item_cascade = _get_option_key(option, path_keys, 'cascade')  # print('查找父option的cascade，直到找到根option')
        if type(item_cascade) == int:
            item_cascade -= len(path_keys)

    if type(item_cascade) != int:
        item_cascade = 0

    item_recursion_value = _get_option_key(option, path_keys, 'recursion_value')

    item_getattrs = _get_option_key(option, path_keys, 'getattrs')
    if item_getattrs:
        attrs = item_getattrs(ins, item_cascade, option, path_keys)
    else:
        attrs = ins.__dict__

    item_include = item_option.get('include', [])
    item_exclude = item_option.get('exclude', [])

    if isfunction(item_include):
        has_item_include = 'func'
    else:
        has_item_include = len(item_include) > 0

    if isfunction(item_exclude):
        has_item_exclude = 'func'
    else:
        has_item_exclude = len(item_exclude) > 0

    result = {}

    for key, value in attrs.items():

        if has_item_include is True and key not in item_include:
            continue
        elif has_item_include == "func" and item_include(ins, key) is not True:
            continue

        if has_item_exclude is True and key in item_exclude:
            continue
        elif has_item_exclude == "func" and item_exclude(ins, key) is True:
            continue

        _value = None
        if isinstance(value, list):
            _value = []
            _isinstance = False
            for item in value:
                if hasattr(item, '__dict__'):
                    _isinstance = True
                    if item_cascade > 0:
                        if item not in path_items:
                            _value.append(_ins_to_dict(item, option, path_items, path_keys + [key]))
                        elif item_recursion_value is not None:
                            _value.append(item_recursion_value)
                else:
                    _value.append(item)
            if _isinstance is True and len(_value) == 0:
                _value = None
        elif hasattr(value, '__dict__'):
            if item_cascade > 0:
                if value not in path_items:
                    _value = _ins_to_dict(value, option, path_items, path_keys + [key])
                elif item_recursion_value is not None:
                    _value = item_recursion_value
        else:
            _value = value

        if _value is not None:
            result[key] = _value
    return result


def _get_option_key(option, key_list, key):
    """
    查找父option的cascade，直到找到根option
    """
    if len(key_list) == 0:
        return option.get(key)

    item_key = '.'.join(key_list)
    item_option = option.get(item_key)
    if isinstance(item_option, dict):
        if key in item_option:
            return item_option.get(key)
    return _get_option_key(option, key_list[:-1], key)



import json

from flask import request

from ..log import flaskz_logger, get_log_data
from ..models import model_to_dict, query_multiple_model, ModelMixin
from ..utils import get_list, is_dict, create_response, get_pss


def init_model_rest_blueprint(model_cls, api_blueprint, url_prefix, module, routers=None, to_json_option=None, multiple_option=None):
    """
    Append rest api route to the api blueprint for the specified model class.
    :param to_json_option: The option used to return json, fields/cascade. ex){'cascade': 1}
    :param module: The module name,ex) role/menu
    :param url_prefix: Api url
    :param api_blueprint:
    :param model_cls: The specified model class.
    :param routers: The api router list, ex)['query', 'query_pss', 'add', 'update', 'delete']
    :param multiple_option:
    :return:
    """
    if to_json_option is None:
        to_json_option = {}

    # from ..sys_mgmt import log_operation

    # permission_required_ = current_app.model_rest_manager._permission_required
    # print(permission_required_)

    class_name = model_cls.get_class_name()
    routers = get_list(routers, ['query', 'query_pss', 'query_multiple', 'add', 'update', 'upsert', 'delete'])

    if 'add' in routers:
        @api_blueprint.route(url_prefix + '/', methods=['POST'])
        @rest_permission_required(module, 'add')
        @gen_route_method('add', url_prefix)
        def add():
            """
            add model data
            :return:
            """
            request_json = request.json
            req_log_data = json.dumps(request_json)

            result = model_cls.add(request_json)
            res_data = model_to_dict(result[1], to_json_option)

            res_log_data = get_log_data(res_data)
            log_operation(module, 'add', result[0], req_log_data, res_log_data)
            flaskz_logger.info(get_rest_log_msg('Add {} data'.format(class_name), req_log_data, result[0], res_log_data))

            return create_response(result[0], res_data)

    if 'delete' in routers:
        @api_blueprint.route(url_prefix + '/<did>', methods=['DELETE'])
        @rest_permission_required(module, 'delete')
        @gen_route_method('delete', url_prefix)
        def delete(did):
            """
            delete model data by id
            :return:
            """
            result = model_cls.delete(did)
            res_data = model_to_dict(result[1], to_json_option)

            res_log_data = get_log_data(res_data)
            log_operation(module, 'delete', result[0], did, res_log_data)
            flaskz_logger.info(get_rest_log_msg('Delete {} data'.format(class_name), did, result[0], res_log_data))

            return create_response(result[0], res_data)

    if 'update' in routers:
        @api_blueprint.route(url_prefix + '/', methods=['PATCH'])
        @rest_permission_required(module, 'update')
        @gen_route_method('update', url_prefix)
        def update():
            """
            Update the model data by the json.
            :return:
            """
            request_json = request.json
            req_log_data = json.dumps(request_json)

            result = model_cls.update(request_json)
            res_data = model_to_dict(result[1], to_json_option)

            res_log_data = get_log_data(res_data)
            log_operation(module, 'update', result[0], req_log_data, res_log_data)
            flaskz_logger.info(get_rest_log_msg('Update {} data'.format(class_name), req_log_data, result[0], res_log_data))

            return create_response(result[0], res_data)

    # if 'update' in routers:
    #     @api_blueprint.route(url_prefix + '/', methods=['PUT'])
    #     @rest_permission_required(module, 'update')
    #     @gen_route_method('update_put', url_prefix)
    #     def update():
    #         """
    #
    #         pass

    if 'upsert' in routers:
        @api_blueprint.route(url_prefix + '/upsert/', methods=['POST'])
        @rest_permission_required(module, 'upsert')
        @gen_route_method('upsert', url_prefix)
        def update():
            """
            Update the model data by the json.
            :return:
            """
            request_json = request.json
            req_log_data = json.dumps(request_json)

            if request_json.get('id'):
                action = "update"
                result = model_cls.update(request_json)
            else:
                action = "add"
                result = model_cls.add(request_json)

            res_data = model_to_dict(result[1], to_json_option)
            res_log_data = get_log_data(res_data)
            log_operation(module, action, result[0], req_log_data, res_log_data)
            flaskz_logger.info(get_rest_log_msg('Upsert {} data'.format(class_name), req_log_data, result[0], res_log_data))

            return create_response(result[0], res_data)

    if 'query' in routers:
        @api_blueprint.route(url_prefix + '/', methods=['GET'])
        @rest_permission_required(module)
        @gen_route_method('query', url_prefix)
        def query():
            """
            Query all the model data
            :return:
            """
            result = model_cls.query_all()
            res_data = model_to_dict(result[1], to_json_option)

            flaskz_logger.debug(get_rest_log_msg('Query all {} data'.format(class_name), None, result[0], res_data))
            return create_response(result[0], res_data)

    if 'query_pss' in routers:
        @api_blueprint.route(url_prefix + '/query_pss/', methods=['GET', 'POST'])
        @rest_permission_required(module)
        @gen_route_method('query_pss', url_prefix)
        def query_pss():
            """
            Query data by search, pagination and sort condition.
            :return:
            """
            request_json = request.json
            req_log_data = json.dumps(request_json)

            result = model_cls.query_pss(get_pss(model_cls, request_json))
            res_data = result[1]
            if result[0] is True:
                res_data['data'] = model_to_dict(res_data['data'], to_json_option)

            flaskz_logger.debug(get_rest_log_msg('Query pss {} data'.format(class_name), req_log_data, result[0], res_data))
            return create_response(result[0], res_data)

    if multiple_option and 'query_multiple' in routers:
        @api_blueprint.route(url_prefix + '/query_multiple/', methods=['GET'])
        @rest_permission_required(module)
        @gen_route_method('query_multiple', url_prefix)
        def query_multiple():
            """
            Query multiple model data
            ex)
            multiple_option={
                'users': User,
                'roles': {
                    'model_cls': Role,
                    'option': {
                        'includes': ['id', 'name']
                    }
                }
            }
            :return:
            """
            model_cls_list = []
            multiple_list = []
            for key in multiple_option:
                item = multiple_option[key]
                item_option = {}
                m_cls = None
                if is_dict(item):  # 'roles': {'model_cls': Role,'option': {'includes': ['id', 'name']}}
                    m_cls = item.get('model_cls')
                    item_option = item.get('option')
                elif isinstance(item, type) and issubclass(item, ModelMixin):  # 'users': User,
                    m_cls = item
                    item_option = {}
                if m_cls and issubclass(m_cls, ModelMixin):
                    model_cls_list.append(m_cls)
                    multiple_list.append({
                        'field': key,
                        'option': item_option
                    })

            result = query_multiple_model(*model_cls_list)
            if result[0] is False:
                success = False
                res_data = model_to_dict(result[1])
            else:
                success = True
                res_data = {}
                for index, item in enumerate(multiple_list):
                    res_data[item.get('field')] = model_to_dict(result[index], item.get('option'))
            flaskz_logger.debug(get_rest_log_msg('Query multiple {} data'.format(class_name), None, success, res_data))
            return create_response(success, res_data)
    return api_blueprint


# def _log_operation(*args, **kwargs):
#     logging_callback = get_current_model_rest_manager_callback('logging_callback')
#     if logging_callback:
#         logging_callback(*args, **kwargs)


from ._mgmt import *
from ._util import *



from flask import request
from sqlalchemy import desc, asc

from ._common import get_dict, is_str, is_dict

__all__ = ['get_remote_addr', 'is_ajax', 'get_pss']


def get_remote_addr():
    """
    Get the remote ip address.
    :return:
    """
    return request.environ.get('HTTP_X_REAL_IP', request.remote_addr)


def is_ajax():
    """
    Check if the request is an ajax request.
    :return:
    """
    return request.headers.get('X-Requested-With') == 'XMLHttpRequest'


def get_pss(cls, pss_config=None):
    """
    Get the search, sort and paging information.
    :param cls:
    :param pss_config:
    :return:
    """
    pss_config = get_dict(pss_config)
    search = pss_config.get('search') or {}
    sort = pss_config.get('sort') or {}
    page = pss_config.get('page') or {}
    # --------------------search--------------------
    ands = []
    ors = []
    search_like = search.pop('like', None)
    like_columns = getattr(cls, 'like_columns', None)
    likes = []
    if search_like and like_columns is not None:
        search_like = str(search_like).strip()
        if search_like != '':
            for col in like_columns:
                if is_str(col):
                    col_field = col
                else:
                    col_field = cls.get_column_field(col)
                likes.append(col_field + " like '%" + search_like + "%'")

    _ands = search.pop('_ands', None)
    if _ands:
        for key in _ands:
            _append_item(ands, key, _ands[key])

    _ors = search.pop('_ors', None)
    if _ors:
        for key in _ors:
            _append_item(ors, key, _ors[key])

    for key in search:
        _append_item(ands, key, search[key])
    # --------------------page--------------------
    offset = page.get('offset') or page.get('skip') or 0
    limit = page.get('limit') or page.get('size') or 100000

    # --------------------query--------------------
    sort_field = sort.get('field')  #
    order = sort.get('order')

    if order and sort_field:
        order = str(order).strip().lower()
        if order in ['desc', 'descend', 'descending']:
            order = desc(sort_field)  # default is desc.
        else:
            order = asc(sort_field)
    elif sort_field:
        if is_str(sort_field):
            order = asc(sort_field)
        else:
            order = sort_field
    else:
        order = None

    return {
        'filter_ands': ands,
        'filter_ors': ors,
        'filter_likes': likes,
        'offset': offset,
        'limit': limit,
        'order': order,
    }


def _append_item(items, key, value):
    if value is not None:
        if is_str(value):
            value = value.strip()
            if value != '':
                val_arr = value.split('||')
                for op_v in val_arr:
                    items.append(key + "='" + op_v + "'")
        elif is_dict(value):
            for operator, op_v in value.items():
                items.append(key + operator + str(op_v))
        else:
            items.append(key + '=' + str(value))
    return items



from sqlalchemy.exc import IntegrityError

from ._base import BaseModelMixin
from .. import res_status_codes
from ..log import flaskz_logger
from ..utils import is_str, is_dict

__all__ = ['ModelMixin']


class ModelMixin(BaseModelMixin):
    # -------------------------------------------add-------------------------------------------
    @classmethod
    def get_add_data(cls, data):
        """
        Return the added json data. By default, return the json data of the client
        Rewrite to customize the added data

        :param data: The data to be added
        :return:
        """
        return data

    @classmethod
    def before_add(cls, data):
        """
        The callback before adding data.
        If the return value is not True, the adding process will be terminated and the result will be returned to the client.
        :param data: The data to be added
        :return: True|Error Message
        """
        return True

    @classmethod
    def after_add(cls, data, instance, before_result):
        """
        The callback after adding data.

        :param data: The data to be added
        :param instance: The result model instance, None if add fails.
        :param before_result: The result of the before_add callback
        :return:
        """
        pass

    @classmethod
    def add(cls, data):
        """
         Add the data to the db.
         Returns a tuple.
           --the first value represents whether the add operation was successful, True/False.
           --the second value represent the result of the add operation, reason/instance.

        :param data:
        :return:
        """
        instance = None
        try:
            data = cls.get_add_data(data)
            check_result = cls.check_add_data(data)
            if check_result is not True:
                return False, check_result

            before_result = cls.before_add(data)
            try:
                if before_result is True:
                    instance = cls.add_db(data)
            finally:
                cls.after_add(data, instance, before_result)  # after_add execute regardless of whether the addition is successful
        except Exception as e:
            print(e)
            flaskz_logger.exception(e)
            return False, res_status_codes.db_add_err

        if before_result is not True:
            return False, before_result  # ex)db_data_already_exist

        return True, instance

    # -------------------------------------------update-------------------------------------------
    @classmethod
    def get_update_data(cls, data):
        """
        Return the updated json data. By default, return the json data of the client
        Rewrite to customize the updated data

        :param data: The updated json data
        :return:
        """
        return data

    @classmethod
    def before_update(cls, data):
        """
        The callback before updating data.
        If the return value is not True, the update process will be terminated and the result will be returned to the client.

        :param data:
        :return:
        """
        return True

    @classmethod
    def after_update(cls, data, instance, before_result):
        """
        The callback after updating data.

        :param data:
        :param instance:
        :param before_result:
        :return:
        """
        pass

    @classmethod
    def update(cls, data):
        """
        Update the data to the db.
        Returns a tuple.
           --the first value represents whether the update operation was successful, True/False.
           --the second value represent the result of the update operation, reason/instance.

        :param data:
        :return:
        """
        instance = None
        try:
            data = cls.get_update_data(data)
            check_result = cls.check_update_data(data)
            if check_result is not True:
                return False, check_result

            before_result = cls.before_update(data)
            try:
                if before_result is True:
                    instance = cls.update_db(data)
            finally:
                cls.after_update(data, instance, before_result)
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_update_err

        if before_result is not True:
            return False, before_result  # ex)db_data_not_found

        if is_str(instance):
            return False, instance
        else:
            return True, instance

    # -------------------------------------------delete-------------------------------------------
    @classmethod
    def get_delete_data(cls, pk_value):
        """
        Return the primary key of the data to be deleted. By default, return the primary key of the client data
        Rewrite to customize the deleted data

        :param pk_value: The primary key of the data to be deleted
        :return:
        """
        return pk_value

    @classmethod
    def before_delete(cls, pk_value):
        """
        The callback before deleting data.
        If the return value is not True, the delete process will be terminated and the result will be returned to the client.

        :param pk_value:
        :return:
        """
        return True

    @classmethod
    def after_delete(cls, pk_value, instance, before_result):
        """
        The callback after deleting data.

        :param pk_value:
        :param instance:
        :param before_result:
        :return:
        """
        pass

    @classmethod
    def delete(cls, pk_value):
        """
        Delete the data from the db.
        Returns a tuple.
           --the first value represents whether the delete operation was successful, True/False.
           --the second value represent the result of the delete operation, reason/instance.

        :param pk_value:
        :return:
        """
        if is_dict(pk_value):  # {id:10}
            pk_value = cls._get_pk_value(pk_value)

        instance = None
        try:
            pk_value = cls.get_delete_data(pk_value)
            check_result = cls.check_delete_data(pk_value)
            if check_result is not True:
                return False, check_result

            before_result = cls.before_delete(pk_value)
            try:
                if before_result is True:
                    instance = cls.delete_db(pk_value)
            finally:
                cls.after_delete(pk_value, instance, before_result)
        except IntegrityError as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_data_in_use
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_delete_err

        if before_result is not True:
            return False, before_result

        if is_str(instance):
            return False, instance
        else:
            return True, instance

    # -------------------------------------------query-------------------------------------------
    @classmethod
    def query_all(cls):
        """
        Override the base query_all method and return success flag
        Used in router to return query data
        :return:
        """
        try:
            return True, super().query_all()
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_query_err

    @classmethod
    def query_pss(cls, pss_option):
        """
        Override the base query_pss method and return success flag
        Used in router to return query data
        :param pss_option:
        :return:
        """
        try:
            return True, super().query_pss(pss_option)
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_query_err



"""
Not included by default.
If use it, please install the 'requests' package first
"""
import inspect

import requests
from flask import request
from requests import Session

from ._common import is_dict, slice_str
from .. import res_status_codes
from ..log import flaskz_logger

__all__ = ['api_request', 'forward_request']

requests_kwargs = None


def _get_request_kwargs(kwargs):
    global requests_kwargs
    if requests_kwargs is None:
        requests_kwargs = []
        for key in inspect.signature(Session.request).parameters:
            requests_kwargs.append(key)

    req_kwargs = {}
    for key, value in kwargs.items():
        if key in requests_kwargs:
            req_kwargs[key] = value
    return req_kwargs


def api_request(url, method="GET", url_params=None, base_url="", raw_response=False, **kwargs):
    """
    Request an api

    :param url:
    :param method:
    :param url_params:
    :param raw_response:
    :param base_url:
    :param kwargs:
    :return:
    """

    _method = method
    if is_dict(url):
        _url = url.get('url')
        if 'method' in url:
            _method = url.get("method")
    else:
        _url = url

    if base_url:
        _url = base_url + _url

    if is_dict(url_params):
        _url = _url.format(**url_params)

    _method = _method.strip().upper()

    flaskz_logger.debug('Api request:\n   url={url}\n   method={method}\n   data={data}\n   json={json}'.format(**{
        'url': _url,
        'method': _method,
        'data': kwargs.get('data'),
        'json': kwargs.get('json')
    }))

    try:
        res = requests.request(method=_method, url=_url, **_get_request_kwargs(kwargs))
        status_code = res.status_code
        result = res.text
        flaskz_logger.debug('Api request completed:\n   status_code={status_code}\n   result={result}'.format(**{
            'status_code': status_code,
            'result': slice_str(result, 60, 10, '\n......\n'),
        }))
        if raw_response is True:
            return res
        return result
    except Exception as e:
        result = str(e)
        flaskz_logger.exception('Api request failed:\n' + str(e))

    return res_status_codes.api_request_err, result


def forward_request(url, payload=None, raw_response=False, error_code=500, **kwargs):
    """
    Forward the request to other service

    :param error_code:
    :param payload:
    :param raw_response:
    :param url: The forward url
    :return:
    """
    req_kwargs = {'url': url}
    _payload = payload or ['method', 'data', 'json', 'headers', 'cookies']
    for item in _payload:
        req_kwargs[item] = getattr(request, item)
    req_kwargs.update(kwargs)  # kwargs high priority

    url_params = req_kwargs.get('url_params')
    if url_params is None:  # if url_params is none, append request.view_args
        req_kwargs['url_params'] = request.view_args or {}

    req_kwargs['raw_response'] = True

    res = api_request(**req_kwargs)
    if raw_response is True:
        return res

    if type(res) is tuple:
        return res[1], error_code

    return res.text, res.status_code, res.headers.items()



import datetime
import json
import logging
import os
from logging.handlers import TimedRotatingFileHandler

from flask import Flask
from werkzeug.local import LocalProxy

_flaskz_logger: logging.Logger
flaskz_logger = LocalProxy(lambda: _get_app_logger())


def init_log(app):
    """
    Initialize system log configuration
    :param app:
    :return:
    """
    if isinstance(app, Flask):
        app_config = app.config
    else:
        app_config = app
    level = logging.getLevelName(app_config.get('FLASKZ_LOGGER_LEVEL'))
    formatter = logging.Formatter(app_config.get('FLASKZ_LOGGER_FORMAT'))
    global _flaskz_logger
    _flaskz_logger = logging.getLogger('flaskz_logger')
    _flaskz_logger.setLevel(level)
    filename = app_config.get('FLASKZ_LOGGER_FILENAME')

    if filename is not None:
        filepath = os.path.join(app_config.get('FLASKZ_LOGGER_FILEPATH') or os.path.join(os.getcwd(), './syslog'), filename)
        log_handler = TimedRotatingFileHandler(
            encoding='utf-8',
            filename=filepath,
            when=app_config.get('FLASKZ_LOGGER_WHEN'),
            interval=1,
            backupCount=int(app_config.get('FLASKZ_LOGGER_BACKUP_COUNT')))
    else:
        log_handler = logging.StreamHandler()

    log_handler.setLevel(level)
    log_handler.setFormatter(formatter)
    _flaskz_logger.addHandler(log_handler)
    _flaskz_logger.disabled = app_config.get('FLASKZ_LOGGER_DISABLED') is True

    wz_logger = logging.getLogger('werkzeug')
    wz_logger.disabled = app_config.get('FLASKZ_WZ_LOGGER_DISABLED') is True


def _get_app_logger():
    return _flaskz_logger or logging.getLogger('flaskz_logger')


def get_log_data(data):
    return json.dumps(data, default=_log_data_converter)


def _log_data_converter(value):
    if isinstance(value, datetime.datetime):
        return value.__str__()
